#define _GNU_SOURCE
#include <unistd.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <assert.h>

#include <MX370x.nsmap> // this file must be included only once in the project

#include "utils.h"

/** @file autorefresh.c
 *
 * @brief This file is an exemple how one may perform an autorefresh acquisition on a MX3701 module using direct access.
 *
 * @note This software is not interactive. Please configure the sensible variables to fit your needs.
 *
 * Commands are sent to the module on top of the SOAP protocol.
 * We use the tool "gsoap" to generate C-code that allows to integrate this protocol transparently in applications.
 * All what developpers have to do is to use the provided API.
 * (This API is detailled in the file soap/doc/html/index.html in this distribution)
 *
 * __Difference between the API description and the actual function calls__
 *
 * The API documentation is accurate but doesn't decrive exactly the prototype of the functions.
 * Anyway transposing a documented function to the actual one is straightforward.
 *
 * Given a function YYY described in the API documentation
 * - the actual function to call are named soap_call_YYY
 * - every function has three more parameters, namelly
 *   - a pointer to a "struct soap" object, that must be initialized first
 * 	 - the SOAP endpoint: a string defining how to access the module. Usually this is a dotted IP address,
 *     but it may as well be a symbolic name ( as long as the DNS resolver knows it )
 *   - a third parameter that is always NULL on our case
 *
 *  Note that the second parameter needs to be given only to the first SOAP call.
 *  As the connection is already established afterwhile, NULL can be used then.
 *
 * This note applies only to function prototypes. Datatypes are exactly described in the API doc.
 */

/* BEGIN OF SOURCE CODE */

//-----------------------------------------------------------------------------------------
/* PARAMETERIZATION OF THE EXAMPLE  */
//-----------------------------------------------------------------------------------------

/* store a pointer to the IP address*/
char * ip = NULL;

//-----------------------------------------------------------------------------------------
/* please see utils.h */
int DUMMY = 0;

//-----------------------------------------------------------------------------------------
/** this is the parameter used to configure an acquisition
 * They are passed to the function soap_call_MX370x__TransducerInitAndStartAutoRefresh()
 *
 */
struct
{
xsd__unsignedLong ulTransducerSelection;
xsd__unsignedLong ulChannelMask;
xsd__unsignedLong ulAverageMode;
xsd__unsignedLong ulAverageValue;
xsd__unsignedLong ulDivisionFactor;
xsd__unsignedLong ulTriggerAction;
xsd__unsignedLong ulHardwareTriggerCount;
xsd__unsignedLong ulHardwareTriggerFilterTime;
xsd__unsignedLong ulByTriggerNbrOfSeqToAcquire;
xsd__unsignedLong ulOption1;
xsd__unsignedLong ulOption2;
xsd__unsignedLong ulOption3;
xsd__unsignedLong ulOption4;
} params =
{
.ulTransducerSelection=3, /**< Transducer type selection */
.ulChannelMask=3, /**< Mask of the channel to acquire by the auto refresh (1 bit = 1 Channel) */
.ulAverageMode=0, /**< Set the average mode :  0 : not used 1 : average per Sequence 2 : average per channel */
.ulAverageValue=0, /**< Set the average value only used, when average is used - 0 : average not used - max value : 255  */
.ulDivisionFactor=5, /**< Division factor  min : 5, max 255  */
.ulTriggerAction=0, /**<
 Hardware Trigger D0 - D7
 Bit 0,1,2,3 : Define the trigger mode
 0000 : Trigger disabled
 0001 : One shot trigger
 0010 : Sequence trigger
 Bit 6,7 : define the active front (Only if hardware trigger selected)
 01 : rising front (Only if hardware trigger selected)
 10 : falling front (Only if hardware trigger selected)
 11 : Both front (Only if hardware trigger selected)
 Software Trigger : D8-D15
 Bit 8,9,10,11 : Define the trigger mode
 0000 : trigger disabled
 0001 : One shot trigger
 0010 : Sequence trigger
 */
.ulHardwareTriggerCount=0, /**< Define the number of trigger events before the action occur
 0 or 1 : all trigger event start the action
 max value : 65535 */
.ulHardwareTriggerFilterTime=0, /**< filter time for the hardware trigger in microseconds
 max value : 65535 */
.ulByTriggerNbrOfSeqToAcquire=0, /**< define the number of sequence to acquire by each trigger event */
.ulOption1=0, /**< Reserved */
.ulOption2=0, /**< Reserved */
.ulOption3=0, /**< Reserved */
.ulOption4=0, /**< Reserved */
};

//-----------------------------------------------------------------------------------------
/* END OF PARAMETERIZATION */
//-----------------------------------------------------------------------------------------

//-----------------------------------------------------------------------------------------
/** Translates error code returned by TransducerInitAndStartAutoRefresh() to a string.
 *
 * @param[in] soapenv The current SOAP environment
 * @param[in] Response The parameter the
 *
 * @return a dynamically allocated string the function TransducerInitAndStartAutoRefresh() had received
 *
 * @note The string returned is dynamically malloc()ated and must be free()d after its use.
 */
static const char* get_TransducerInitAndStartAutoRefresh_error_message(struct soap* soapenv, struct MX370x__Response* Response)
{
	switch(Response->iReturnValue)
	{
		case 0:
			return strdup("remote function performed OK");
		case -1:
			/* means an system error occured */
			{
				struct MXCommon__ByteArrayResponse Response1;
				assert_gsoap(soapenv,soap_call_MXCommon__Strerror(soapenv,NULL, NULL,Response->syserrno, &Response1) );
				return strdup(Response1.sArray.__ptr);
			}
		case -2:
		 	return strdup("Transducer selection error");
		case -3:
		 	return strdup("Channel mask error : can not be null");
		case -4:
		 	return strdup("Channel mask error");
		case -5:
		 	return strdup("Average mode error");
		case -6:
		 	return strdup("Average value error");
		case -7:
		 	return strdup("Division factor error");
		case -8:
		 	return strdup("Trigger action : hardware trigger 2 different mode cannot be simultaneously be activated");
		case -9:
		 	return strdup("Trigger action : front definition error");
		case -10:
		 	return strdup("Trigger action : software trigger 2 different mode cannot be simultaneously be activated");
		case -11:
		 	return strdup("Hardware Trigger count error");
		case -12:
		 	return strdup("Hardware Trigger filter time error");
		case -13:
		 	return strdup("By trigger number of sequence to acquire error");
		case -100:
		 	return strdup("TransducerInit kernel function error");
		case -101:
		 	return strdup("InitConvertTimeDivisionFactor kernel function error");
		case -102:
		 	return strdup("SetAutoRefreshAverageValue kernel function error");
		case -103:
		 	return strdup("InitDigitalInputFilter kernel function error");
		case -104:
		 	return strdup("InitEnableDisableHardwareTrigger kernel function error");
		case -105:
			return strdup("StartAutoRefresh kernel function error");
		default:
			return(strdup("unknown error code"));
	}
}

//-----------------------------------------------------------------------------------------
/** signal handler : stops the current acquisition and cleanup the SOAP connection.
 *
 */
static void do_signal (int sig)
{
	struct soap soapenv;
	mx_soap_init(&soapenv);
	char endpoint[30];

	printf("signal received:%s\n",strsignal(sig));
	printf("stopping current acquisition\n");

	/* generates the endpoint string
	 * SOAP "endpoints" are of the form NAME:PORT -
	 * PORT is in the case of the SOAP command server: 5555
	 */
	if ( snprintf(endpoint,sizeof(endpoint),"%s:5555",ip) < 0)
	{
		fprintf(stderr,"error: %s:%d!\n",__FUNCTION__,__LINE__);
		exit(1);
	}
	/* note: the soapenv variable should have been correctly initialised before the signal handler was set */
	{
		struct MX370x__Response Response;
		if ( ! (SOAP_OK==soap_call_MX370x__TransducerStopAndReleaseAutoRefresh(&soapenv, endpoint, NULL, &DUMMY, &Response)) )
		{
			fprintf(stderr,"error while stopping acquisition: ");
			soap_print_fault(&soapenv, stderr);
		}
		mx_soap_clean(&soapenv);
	}

	exit(1);
}

//-----------------------------------------------------------------------------------------
/** For all signals set the handler do_signal.
 *
 * This allows to handle CTRL+C nicely
 *
 * @note This function must be called after the first call to a function of the soap_call() familly.
 */
static void setup_signal(void)
{
	struct sigaction action;
	/* catch all signal */
	sigfillset(&action.sa_mask); /* all other signal will be blocked when handling the one received */
	action.sa_handler = do_signal;
	action.sa_flags = 0;
	{
		int i;
		for (i=1; i< NSIG; i++)
		{
			if (SIGKILL==i) /* this signal can not be blocked */
				continue;
			if (SIGSTOP==i) /* this signal can not be blocked */
				continue;
			sigaction(i,&action,NULL); /* TODO: manage error */
		}
	}
}

//-----------------------------------------------------------------------------------------

/**
 * Function called when the parameters of the sample are wrong
 * The sample need the IP address of the system as parameter
 */
static void usage()
{
	fprintf(stderr, "Usage: ./autorefresh IP_ADDRESS TransducerIndex \n");
	fprintf(stderr, "TransducerIndex: Index of the transducer you want to use (see retrieveinfo sample -> Selection Index).\n");
	exit(1);
}

//-----------------------------------------------------------------------------------------
int main(int argc, char * argv[])
{
	/** this variable contains the current SOAP connection */
	struct soap soapenv;

	char endpoint[30];
	int clientfd = -1;
	unsigned int number_of_channels = 0;

	/* check the parameters */
	if ( argc != 3 )
		usage();

	/* store a pointer to the IP */
	ip = argv[1];

	/* get the transducer index */
	params.ulTransducerSelection = atol(argv[2]);

	/* data initialisation */
	memset(endpoint	,0,sizeof(endpoint));
	mx_soap_init(&soapenv);

	/* generates the endpoint string
	 * SOAP "endpoints" are of the form NAME:PORT -
	 * PORT is in the case of the SOAP command server: 5555
	 */
	if ( snprintf(endpoint,sizeof(endpoint),"%s:5555",argv[1]) < 0)
	{
		fprintf(stderr,"error: %s:%d!\n",__FUNCTION__,__LINE__);
		exit(1);
	}

	/* connect to data server (if no acquisition was started, we should not receive any data) */
	clientfd = datasrv_connect(argv[1]);
	/* retrieve the number of channels on the module (4,8 or 16)
	 *
	 * @note On upcoming version of the firmware there will be a SOAP function dedicated to
	 * this task.
	 *
	 */
	{
		struct MXCommon__ByteArrayResponse response;

		assert_gsoap(&soapenv, soap_call_MXCommon__GetModuleType(&soapenv,endpoint , NULL, &DUMMY,&response) );
		/* extract the number of channels from the "type" string */
		assert( ! parse_number_of_channels (response.sArray.__ptr, &number_of_channels) );
	}

	/* send request to start the acquisition, using parameters set in the "params" global variable. */
	{
		struct MX370x__Response Response; // this is a general type for communicating with the MX module
		assert_gsoap(	&soapenv,
						soap_call_MX370x__TransducerInitAndStartAutoRefresh (
															&soapenv,
															endpoint,
															NULL,
															params.ulTransducerSelection,
															params.ulChannelMask,
															params.ulAverageMode,
															params.ulAverageValue,
															params.ulDivisionFactor,
															params.ulTriggerAction,
															params.ulHardwareTriggerCount,
															params.ulHardwareTriggerFilterTime,
															params.ulByTriggerNbrOfSeqToAcquire,
															params.ulOption1,
															params.ulOption2,
															params.ulOption3,
															params.ulOption4,
															&Response) );
		/* manage errors */
		if ( Response.iReturnValue )
		{
			char * msg = get_TransducerInitAndStartAutoRefresh_error_message(&soapenv,&Response);
			fprintf(stderr,"error at %s:%d\nmessage: %s\n", __FUNCTION__,__LINE__,msg);
			if (msg) free(msg);
			exit(1);
		}
	}

  	/* from now on, we want the acquisition to be stopped on program termination */
  	setup_signal();
  	/* now we should use abort() instead of exit() */

  	{
  		/* buffer that will contain the value */
		long unsigned int * buffer = NULL;

		/* Compute the amount of data to read at one time, in bytes
		 *
		 * a value is a 4-bytes integer. the number of value is the number of channel to acquire +1
		 * the first value is the counter
		 *
		 */
		int packetsize = 4; /* size of data packet including counter */

		{
			int i;
			for (i = 0; i < number_of_channels; i ++)
			{
				if (params.ulChannelMask & (1<<i))
					packetsize += 4;
			}
		}

		/* allocate buffer */
		if ( ! (buffer = malloc(packetsize) ) )
		{
			perror("malloc");
			exit(1);
		}

		/* Release the memory allocated by the gSOAP calls */
		soap_destroy(&soapenv);
		soap_end(&soapenv);

		/* fetch data sent by the data server and print it out */
		for (;;)
		{
			if ( recv(clientfd,buffer,packetsize,0) < 1 )
			{
				/* two cases:
				 * - the peer has performed an orderly shutdown. ( returned 0)
				 * - another error happened ( returned -1)
				 */
				perror("recv");
				abort();
			}
			/* there is data in the buffer. print it out */
			{
				int i;
				printf("Counter: %lu",buffer[0]);
				for (i=1; i< (packetsize/4); i++)
				{
					printf(" %d:%lX",i,buffer[i]);
				}
				printf("\n");
			}

		}
  	}

	printf("OK\n");
	return 0;
}
