#include <unistd.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <MX370x.nsmap> // this file must be included only once in the project
#include <assert.h>

#include "utils.h"

/** @file retrieveinfo.c
 *
 * @brief This file is an exemple how one may use SOAP commands to retrieve information form a given module.
 *
 * The information retrieved are the number of channels and the database of transducer type
 * currently configured in the module.
 *
 * @note This software is not interactive. Please configure the sensible variables to fit your needs.
 *
 */

//-----------------------------------------------------------------------------------------
/* PARAMETERIZATION OF THE EXAMPLE  */
//-----------------------------------------------------------------------------------------

/* store a pointer to the IP address*/
char * ip = NULL;

//-----------------------------------------------------------------------------------------
/* END OF PARAMETERIZATION */
//-----------------------------------------------------------------------------------------

//-----------------------------------------------------------------------------------------
/* please see utils.h */
int DUMMY = 0;

//-----------------------------------------------------------------------------------------
char *temperaturestr[] =
{
 [0]= "TEMPERATURE_INITIAL", /* Temperature not ready */
 [1]= "TEMPERATURE_TOOLOW", /* Temperature too low !  */
 [2]= "TEMPERATURE_LOW", /* Temperature under the min warning value  */
 [3]= "TEMPERATURE_NOMINAL", /* Temperature in the nominal range */
 [4]= "TEMPERATURE_HIGH", /* Temperature over the max warning value */
 [5]= "TEMPERATURE_TOOHIGH",
};

//-----------------------------------------------------------------------------------------

/**
 * Function called when the parameters of the sample are wrong
 * The sample need the IP address of the system as parameter
 */
static void usage()
{
	fprintf(stderr, "Usage: ./autorefresh IP_ADDRESS\n");
	exit(1);
}


//-----------------------------------------------------------------------------------------
int main(int argc, char * argv[])
{
	char endpoint[30];
	memset(endpoint	,0,sizeof(endpoint));

	/* this structure  the current SOAP connection  */
	struct soap soapenv;

	/* check the parameters */
	if ( argc != 2 )
		usage();

	/* store a pointer to the IP */
	ip = argv[1];

	mx_soap_init(&soapenv);

	/* generates the endpoint string
	 * SOAP "endpoints" are of the form NAME:PORT - PORT is in the case of the SOAP command server 5555
	 */
	if ( snprintf(endpoint,sizeof(endpoint),"%s:5555",argv[1]) < 0)
	{
		fprintf(stderr,"error: %s:%d!\n",__FUNCTION__,__LINE__);
		exit(1);
	}

	/* ask the module for its type and number of channels */
	{
		unsigned int number_of_channels = 0;
		struct MXCommon__ByteArrayResponse response;

		assert_gsoap(&soapenv, soap_call_MXCommon__GetModuleType(&soapenv,endpoint , NULL, &DUMMY,&response) );
		printf("returned string: %s\n",response.sArray.__ptr);

		/* get the number of channels */
		assert( ! parse_number_of_channels (response.sArray.__ptr, &number_of_channels) );
		printf("this module has %d channels\n",number_of_channels);
	}

	/* who is currently connected ? */
	{
		struct MXCommon__ByteArrayResponse Response;
		assert_gsoap(&soapenv,soap_call_MXCommon__GetClientConnections(  &soapenv,NULL, NULL, &DUMMY,&Response ));
		assert_soap_command(&soapenv,&(Response.sResponse) );
		printf("current connections: %s\n\n", Response.sArray.__ptr);
	}

	/* current temperature ? */
	{
		struct MXCommon__GetModuleTemperatureValueAndStatusResponse Response;
		assert_gsoap(&soapenv,soap_call_MXCommon__GetModuleTemperatureValueAndStatus(  &soapenv,NULL, NULL, 0 /*unused*/, &Response));
		assert_soap_command(&soapenv,&(Response.sResponse) );
		printf("current temperature: %lf\n",Response.dTemperatureValue);
		printf("temperature status is %s\n", temperaturestr[Response.ulTemperatureStatus]);
	}
	/* display transducers database */
	{
		struct MX370x__unsignedlongResponse Response;
		struct MX370x__TransducerGetTypeInformationResponse TypeInfoResponse;

		/* Get the number of transducer type */
		assert_gsoap(&soapenv, soap_call_MX370x__TransducerGetNbrOfType(&soapenv, NULL, NULL, &DUMMY, &Response) );
		printf("\nNumber of transducer types: %ld\n", Response.ulValue);

		/* go through the database and fetch data */
		{
			int i;
			for (i = 0; i < Response.ulValue; i ++)
			{
				/* Get the transducer information */
				assert_gsoap(&soapenv,soap_call_MX370x__TransducerGetTypeInformation (&soapenv,NULL, NULL, i, &TypeInfoResponse) );

				if (TypeInfoResponse.iReturnValue == 0)
				{
					printf("\n");
					printf("Transducer Index: %d\n", i);
					/* use the value below to identify a transducer type
					 * in call of functions of the family "TransducerInitAndStart"
					 */
					printf("Selection Index: %ld\n", TypeInfoResponse.ulTransducerSelectionIndex);
					printf("Name: %s\n", TypeInfoResponse.pcName);
					printf("Calibration status: %ld\n", TypeInfoResponse.ulCalibrationStatus);
					printf("Type value: %ld\n", TypeInfoResponse.ulType);
					printf("Frequency: %ld Hz\n", TypeInfoResponse.ulFrequency);
					printf("Impedance: %ld Ohm\n", TypeInfoResponse.ulImpedance);
					printf("Nominal Voltage: %lf V\n", TypeInfoResponse.dVeff);
					printf("Sensibility: %lf mV/V/mm\n", TypeInfoResponse.dSensibility);
					printf("Range: +/- %lf mm\n", TypeInfoResponse.dRange);
				}

				/* Release the memory allocated by the gSOAP calls */
				soap_destroy(&soapenv);
				soap_end(&soapenv);
			}
		}
	}
	return 0;
}
